using Godot;
using System;

public class Character : KinematicBody2D
{

    [Export]
    private float MaxHealth = 100f;
    private float CurrentHealth;

    [Export]
    protected float RegenerationPerSecond = 0f;
    protected float RPSAddition = 0f; // RPS = RegenerationPerSecon
    protected float RPSRelativeBefore = 0f; // RPS = RegenerationPerSecond
    protected float RPSRelativeAfter = 0f; // RPS = RegenerationPerSecond

    private float _deltaDamage = 0f; // time since last damage

    [Export]
    private bool IsRegenerationOn = false;

    [Export]
    public bool IsAlive = true;

    [Export]
    public bool IsGravityEnabled = true;

    [Export]
    private float Gravity = 3000;

    [Export]
    private float MaxFallSpeed = 1250;

    [Export]
    private float JumpHeight = 15000;

    [Export]
    private float JumpSpeed = 50000;

    [Export]
    private float MinJumpHeight = 0;

    private bool _currentlyJumping = false;
    private float _deltaJumpStart = 0;

    [Export]
    public bool IsJumpingEnabled = true;


    private Vector2 Velocity = new Vector2();


    [Export]
    private float MaxRunSpeed = 500;

    private float RunSpeedMultiplier = 1;

    [Export]
    private float Acceleration = 40;

    [Export]
    private float Deceleration = 35;

    [Export]
    private float AirAcceleration = 30;

    [Export]
    private float AirDeceleration = 10;

    [Export]
    public bool IsRunningEnabled = true;

    protected bool ShallRunRight = false;
    protected bool ShallRunLeft = false;
    protected bool ShallJump = false;

    public float GetRunSpeed()
    {
        return MaxRunSpeed * RunSpeedMultiplier;
    }

    public float GetAcceleration()
    {
        return Acceleration;
    }

    public float GetAirAcceleration()
    {
        return AirAcceleration;
    }

    public float GetDeceleration()
    {
        return Deceleration;
    }

    public float GetAirDeceleration()
    {
        return AirDeceleration;
    }

    public float GetGravity()
    {
        return this.Gravity;
    }
    public float SetMaxHealth(float value)
    {
        this.MaxHealth = value;
        return value;
    }

    public float SetCurrentHealth(float value)
    {
        this.CurrentHealth = value;
        return value;
    }

    public float GetMaxHealth()
    {
        return this.MaxHealth;
    }

    public float GetCurrentHealth()
    {
        return this.CurrentHealth;
    }

    public float GetDeltaDamage()
    {
        return _deltaDamage;
    }

    public Vector2 GetVelocity()
    {
        return Velocity;
    }

    public void SetVelocityX(float value)
    {
        this.Velocity.x = value;
    }

    public void SetVelocityY(float value)
    {
        this.Velocity.y = value;
    }

    public bool IsStepping()
    {
        return this._isStepping;
    }

    public bool IsJumping()
    {
        return _currentlyJumping;
    }

    public void AddVelocity(Vector2 value)
    {
        this.Velocity += value;
    }

    public float AddMaxHealth(float value)
    {
        this.MaxHealth += value;
        return this.MaxHealth;
    }

    public float AddCurrentHealth(float value)
    {
        this.CurrentHealth += value;
        return this.CurrentHealth;
    }

    public float SecureAddCurrentHealth(float value)
    {
        if (this.AddCurrentHealth(value) > this.MaxHealth)
        {
            this.SetCurrentHealth(this.MaxHealth);
        }
        return this.CurrentHealth;
    }

    public float SubMaxHealth(float value)
    {
        this.MaxHealth += value;
        return this.MaxHealth;
    }

    public float SubCurrentHealth(float value)
    {
        this.CurrentHealth -= value;
        return this.CurrentHealth;
    }

    public void Damage(float value)
    {
        if (this.IsAlive)
        {
            if (this.SubCurrentHealth(value) < 0)
            {
                this.SetCurrentHealth(0);
                this.IsAlive = false;
            }
        }
    }
    private float StepVal = 0;
    private float StepSpeed = 1000;
    private bool _isStepping = false;

    private float _deltaStepping = 0;

    public void StepOn(float value)
    {
        if (!_currentlyJumping)
        {
            StepVal += value;
            _isStepping = true;
            _deltaStepping = 0;
        }
    }

    private void CalcStep(float delta)
    {
        if (_isStepping)
        {
            var stepSpeed = StepVal < 0 ? -StepSpeed : StepSpeed;
            if (_deltaStepping >= StepVal / stepSpeed)
            {
                _isStepping = false;
                this.Velocity.y = 0;
                this.StepVal = 0;
            }
            else
            {
                this.Velocity.y = stepSpeed;
            }
            _deltaStepping += delta;
        }
    }

    private void CalcJump(float delta)
    {
        if (!this.ShallJump)
        {
            if (this._deltaJumpStart >= this.MinJumpHeight / this.JumpSpeed)
            {
                this._currentlyJumping = false;
            }
        }
        if (this.IsOnCeiling())
        {
            this._currentlyJumping = false;
            this.Velocity.y = 0;
        }

        if (this._currentlyJumping)
        {
            this._deltaJumpStart += delta;
            this.Velocity.y = -JumpSpeed * delta;
            if (this._deltaJumpStart >= this.JumpHeight / this.JumpSpeed || this.IsOnCeiling())
            {
                this._currentlyJumping = false;
            }
        }
        else
        {
            if (this.IsOnFloor())
            {
                this._currentlyJumping = this.ShallJump;
                if (this.ShallJump)
                {
                    this._deltaJumpStart = 0;
                }
            }
        }
    }

    private void CalcRun(float delta)
    {

        // Acceleration
        var acceleration = IsOnFloor() ? GetAcceleration() : GetAirAcceleration();
        if (ShallRunLeft)
        {
            Velocity.x -= acceleration;
        }
        if (ShallRunRight)
        {
            Velocity.x += acceleration;
        }

        // Deceleration
        if (!(ShallRunLeft ^ ShallRunRight)  // There is no deceleration if the Character is Running. But if he is running at both sides concurrently (so actually not running) there is.
        || (ShallRunLeft && Velocity.x > 0) || (ShallRunRight && Velocity.x < 0))
        {
            var deceleration = IsOnFloor() ? GetDeceleration() : GetAirDeceleration();
            if (Velocity.x > 0)
            {
                Velocity.x -= deceleration;
                if (Velocity.x < 0) // Checking to ensure that there is no acceleration in the other direction
                {
                    Velocity.x = 0;
                }
            }
            else
            {
                Velocity.x += deceleration;
                if (Velocity.x > 0) // Checking to ensure that there is no acceleration in the other direction
                {
                    Velocity.x = 0;
                }
            }
        }

        var runSpeed = GetRunSpeed();
        if (Velocity.x > runSpeed)
        {
            Velocity.x = runSpeed;
        }
        else if (Velocity.x < -runSpeed)
        {
            Velocity.x = -runSpeed;
        }

    }

    private void CalcGravity(float delta)
    {
        if (!_currentlyJumping)
        {
            this.Velocity.y += delta * this.Gravity;
            if (this.Velocity.y > this.MaxFallSpeed)
            {
                this.Velocity.y = this.MaxFallSpeed;
            }
        }
    }

    // Called every frame. 'delta' is the elapsed time since the previous frame.
    public override void _PhysicsProcess(float delta)
    {
        base._PhysicsProcess(delta);


        if (this.IsJumpingEnabled)
        {
            CalcJump(delta);
        }

        if (this.IsGravityEnabled && !_isStepping)
        {
            CalcGravity(delta);
        }

        if (IsRunningEnabled)
        {
            this.CalcRun(delta);
        }

        this.CalcStep(delta);


        MoveAndSlide(Velocity, new Vector2(0, -1));

        if (this.IsOnFloor())
        {
            this.Velocity.y = 0;
        }
    }

    public override void _Process(float delta)
    {
        base._Process(delta);

        _deltaDamage += delta;

        if (this.IsRegenerationOn && this.IsAlive)
        {
            if (this.CurrentHealth < this.MaxHealth)
            {
                var specialMultiplier = CalcSpecialRegenerationMultiplier(GetDeltaDamage());
                if (specialMultiplier != 0)
                {
                    float amountToAdd = (this.RegenerationPerSecond * RPSRelativeBefore +
                        (RegenerationPerSecond + RPSAddition) * (RPSRelativeAfter + 1)) * delta * specialMultiplier;
                    this.SecureAddCurrentHealth(amountToAdd);
                }
            }
        }
    }
    public override void _Ready()
    {
        this.CurrentHealth = MaxHealth;
    }

    // Override this function, if you want to change the amount of regeneration based on the last time the character got damage.
    // deltaDamage is the time since last damage taken.
    // DONT CALL THE FUNCTION OUTSIDE THIS CLASS. THE FUNCTIONALITY IS HANDLED INTERNALLY.
    protected virtual float CalcSpecialRegenerationMultiplier(float deltaDamage)
    {
        return 1f;
    }

}
